from django.db import models
from slackchat.conf import settings


class User(models.Model):
    """
    A Slack user who posts in a slackchat channel.
    """

    api_id = models.CharField(max_length=50)

    first_name = models.CharField(max_length=100, blank=True, null=True)
    last_name = models.CharField(max_length=200, blank=True, null=True)
    email = models.EmailField(blank=True, null=True)
    image = models.ImageField(upload_to=settings.USER_IMAGE_UPLOAD_TO)
    title = models.CharField(max_length=255)

    def __str__(self):
        return '{0} {1}'.format(self.first_name, self.last_name)



from django.db import models


class ChatType(models.Model):
    """
    A type of slackchat.
    """

    name = models.CharField(max_length=255)
    render_to_html = models.BooleanField(
        default=False,
        help_text="Whether to render markdown to HTML in the serializer.")
    kwargs_in_threads = models.BooleanField(
        default=True,
        help_text="Whether users can create kwargs in threads."
    )

    def __str__(self):
        return self.name



from django.urls import include, path
from rest_framework import routers
from rest_framework_swagger.views import get_swagger_view

from .views import Events
from .viewsets import ChannelViewset

router = routers.DefaultRouter()
router.register(r'channel', ChannelViewset)

schema_view = get_swagger_view(title='Slackchat API Docs')

urlpatterns = [
    path('api/', include(router.urls)),
    path('api/docs/', schema_view),
    path('events/', Events.as_view()),
]



# flake8: noqa
from .argument import Argument
from .channel import Channel
from .chat_type import ChatType
from .custom_content_template import CustomContentTemplate
from .keyword_argument import KeywordArgument
from .message import Message
from .reaction import Reaction
from .user import User
from .webhook import Webhook



from rest_framework import serializers
from slackchat.models import Channel

from .message import MessageSerializer
from .user import UserSerializer


class ChannelSerializer(serializers.ModelSerializer):
    chat_type = serializers.SerializerMethodField()
    users = serializers.SerializerMethodField()
    meta = serializers.SerializerMethodField()
    introduction = serializers.SerializerMethodField()
    messages = MessageSerializer(many=True)

    def get_chat_type(self, obj):
        return obj.chat_type.name

    def get_users(self, obj):
        users = {}

        for message in obj.messages.all():
            user = message.user
            serializer = UserSerializer(instance=user)
            if user.api_id not in users:
                users[user.api_id] = serializer.data

            for reaction in message.reactions.all():
                serializer = UserSerializer(instance=reaction.user)
                if reaction.user.api_id not in users:
                    users[reaction.user.api_id] = serializer.data

        return users

    def get_meta(self, obj):
        return {
            "title": obj.meta_title,
            "description": obj.meta_description,
            "keywords": obj.meta_keywords
        }

    def get_introduction(self, obj):
        return obj.get_introduction()

    class Meta:
        model = Channel
        fields = (
            'id',
            'api_id',
            'chat_type',
            'title',
            'image',
            'introduction',
            'meta',
            'publish_path',
            'users',
            'messages',
        )



from django.db import models


class Webhook(models.Model):
    """
    An endpoint that is hit whenever Slackchat objects
    are created or updated.

    Used to signal reserialization should happen for
    a slackchat renderer.
    """
    endpoint = models.URLField(unique=True)
    verified = models.BooleanField(default=False)

    def __str__(self):
        return self.endpoint



from django.db import models


class CustomContentTemplate(models.Model):
    """
    Defines search parameters for finding custom messages
    and a template for how to serialize the message.
    """

    name = models.CharField(max_length=255)
    argument_name = models.SlugField(
        max_length=255,
        blank=True, null=True,
        help_text="Add an argument to the message if search_string \
        matches against a message's content.")
    search_string = models.CharField(
        max_length=255,
        help_text="A regex search string with capture groups.")
    chat_type = models.ForeignKey(
        'ChatType', on_delete=models.CASCADE)
    content_template = models.TextField(
        help_text="A Python format string whose args are the capture groups \
        matched by the search_string."
    )

    def __str__(self):
        return self.name



from datetime import datetime

from django.core.exceptions import ObjectDoesNotExist
from markslack import MarkSlack
from slackchat.conf import settings
from slackchat.exceptions import MessageNotFoundError, UserNotFoundError
from slackchat.models import Channel, KeywordArgument, Message, User

ignored_subtypes = [
    'group_join',
    'file_share',
    'group_archive',
]


def strptimestamp(timestamp):
    return datetime.fromtimestamp(float(timestamp))


def handle_removed(id, event):
    try:
        channel = Channel.objects.get(
            api_id=event.get('channel')
        )
    except ObjectDoesNotExist:
        return
    msg = {
        'user': event.get('previous_message').get('user'),
        'ts': event.get('previous_message').get('ts'),
        'text': event.get('message', {}).get('text', None)
    }
    user, created = User.objects.get_or_create(
        api_id=msg.get('user')
    )

    # If thread
    if event.get('previous_message', {}).get('thread_ts'):
        if not channel.chat_type.kwargs_in_threads:
            return

        thread = event.get('previous_message')

        try:
            original_message = Message.objects.get(
                timestamp=strptimestamp(thread.get('thread_ts'))
            )
        except ObjectDoesNotExist:
            raise MessageNotFoundError(
                '{}'.format(strptimestamp(thread.get('thread_ts')))
            )

        try:
            user = User.objects.get(api_id=thread.get('user'))
        except ObjectDoesNotExist:
            raise UserNotFoundError(thread.get('user'))

        try:
            kwarg = KeywordArgument.objects.get(
                timestamp=strptimestamp(thread.get('ts')),
                message=original_message,
                user=user
            )
        except ObjectDoesNotExist:
            return

        kwarg.delete()

    else:
        Message.objects.get(
            channel=channel,
            timestamp=strptimestamp(msg.get('ts')),
            user=user,
        ).delete()


def handle(id, event):
    user_templates = {
        user.api_id: settings.MARKSLACK_USER_TEMPLATE(user)
        for user in User.objects.all()
    }

    marker = MarkSlack(
        user_templates=user_templates,
        link_templates=settings.MARKSLACK_LINK_TEMPLATES,
        image_template=settings.MARKSLACK_IMAGE_TEMPLATE
    )

    try:
        channel = Channel.objects.get(
            api_id=event.get('channel')
        )
    except ObjectDoesNotExist:
        return

    subtype = event.get('subtype', None)
    if subtype in ignored_subtypes:
        return

    if subtype:
        msg = {
            'user': event.get('previous_message').get('user'),
            'ts': event.get('previous_message').get('ts'),
            'text': event.get('message', {}).get('text', None)
        }
    else:
        msg = {
            'user': event.get('user'),
            'ts': event.get('ts'),
            'text': event.get('text')
        }

    user, created = User.objects.get_or_create(
        api_id=msg.get('user')
    )

    thread_ts = event.get('thread_ts', None) or \
        event.get('message', {}).get('thread_ts', None)
    # If thread
    if thread_ts and (
        event.get('parent_user_id', None) or
        event.get('message', {}).get('parent_user_id', None)
    ):
        if not channel.chat_type.kwargs_in_threads:
            return

        try:
            key, value = msg.get('text').split(': ', 1)
        except ValueError:
            # Don't handle threads that don't follow key-value pattern
            return

        try:
            original_message = Message.objects.get(
                timestamp=strptimestamp(thread_ts)
            )
        except ObjectDoesNotExist:
            raise MessageNotFoundError(
                '{}'.format(strptimestamp(thread_ts))
            )
        KeywordArgument.objects.update_or_create(
            timestamp=strptimestamp(msg.get('ts')),
            message=original_message,
            user=user,
            defaults={
                "key": key,
                "value": value
            }
        )
    else:
        message, created = Message.objects.update_or_create(
            channel=channel,
            timestamp=strptimestamp(msg.get('ts')),
            user=user,
            defaults={
                'text': marker.mark(msg.get('text'))
            }
        )



from rest_framework import viewsets

from .models import Channel
from .serializers import ChannelSerializer


class ChannelViewset(viewsets.ModelViewSet):
    queryset = Channel.objects.all()
    serializer_class = ChannelSerializer
    lookup_field = 'pk'



import uuid

from django.db import models
from django.utils.encoding import escape_uri_path
from django.utils.safestring import mark_safe
from markdown import markdown
from slackchat.conf import settings
from slackclient import SlackClient


class Channel(models.Model):
    """
    A Slack channel that hosts a slackchat.
    """

    def fetch_slack_users():
        client = SlackClient(settings.SLACK_API_TOKEN)
        response = client.api_call("users.list")
        return [
            (m.get('id'), m.get('profile').get('email'))
            for m in response.get("members")
            if not m.get('deleted') and
            m.get('profile', {}).get('email', False)
        ]

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    api_id = models.SlugField(
        max_length=10, null=True, blank=True, editable=False)

    chat_type = models.ForeignKey(
        'ChatType', on_delete=models.PROTECT)

    owner = models.CharField(max_length=200, choices=fetch_slack_users())

    title = models.CharField(max_length=150, blank=True, null=True)

    image = models.ImageField(
        upload_to=settings.CHANNEL_IMAGE_UPLOAD_TO, blank=True, null=True,
        help_text="An image to feature on the rendered Slackchat page.")

    introduction = models.TextField(
        blank=True, null=True,
        help_text="Some introductory paragraph text in markdown syntax.")

    meta_title = models.CharField(
        max_length=300, blank=True, null=True,
        help_text="Title for page meta data.")
    meta_description = models.CharField(
        max_length=300, blank=True, null=True,
        help_text="Description for page meta data.")
    meta_keywords = models.CharField(
        max_length=300, blank=True, null=True,
        help_text="Keywords for page meta data.")

    publish_path = models.CharField(
        max_length=300, blank=True, null=True,
        help_text="A relative path that a renderer may use when \
        publishing the slackchat."
    )

    def save(self, *args, **kwargs):
        if self.publish_path:
            self.publish_path = escape_uri_path(self.publish_path)
        super().save(*args, **kwargs)

    def get_introduction(self):
        if self.chat_type.render_to_html:
            return mark_safe(markdown(self.introduction))
        return self.introduction

    def __str__(self):
        return 'slackchat-{}'.format(self.id.hex[:10])



from django.contrib import admin

from .models import (Argument, Channel, ChatType, CustomContentTemplate,
                     KeywordArgument, Message, Reaction, User)


class ChannelAdmin(admin.ModelAdmin):
    readonly_fields = ('api_id',)
    fieldsets = (
        (None, {
            'fields': ('api_id', 'chat_type', 'owner')
        }),
        ('Publishing', {
            'fields': ('publish_path',)
        }),
        ('Display', {
            'fields': ('image', 'title', 'introduction')
        }),
        ('Meta', {
            'fields': ('meta_title', 'meta_description', 'meta_keywords')
        })
    )


admin.site.register(Argument)
admin.site.register(Channel, ChannelAdmin)
admin.site.register(User)
admin.site.register(CustomContentTemplate)
admin.site.register(Message)
admin.site.register(Reaction)
admin.site.register(KeywordArgument)
admin.site.register(ChatType)



import uuid

import requests

from celery import shared_task
from slackchat.conf import settings
from slackchat.models import Webhook


@shared_task(acks_late=True)
def post_webhook(channel_id):
    data = {
        "token": settings.WEBHOOK_VERIFICATION_TOKEN,
        "type": "update_notification",
        "channel": channel_id,
    }
    for webhook in Webhook.objects.all():
        if webhook.verified:
            requests.post(webhook.endpoint, data=data)


@shared_task(acks_late=True)
def verify_webhook(pk):
    webhook = Webhook.objects.get(pk=pk)
    if not webhook.verified:
        challenge = uuid.uuid4().hex[:10]
        response = requests.post(webhook.endpoint, data={
            "token": settings.WEBHOOK_VERIFICATION_TOKEN,
            "type": "url_verification",
            "challenge": challenge,
        })
        if response.status_code == requests.codes.ok:
            if response.text == challenge:
                webhook.verified = True
                webhook.save()
